@@ -6,6 +6,9 @@ Sriperumbudur (NIPS #22, 2009) and 'A Kernel Two-Sample Test', A. Gretton,
 K. M. Borgwardt, M. J. Rasch, B. Sch\"{o}lkopf and A. Smola (Journal of
 Machine Learning Research #13, 2012).
 """
+import copy
+import itertools
+
 import numpy as np
 
 from enum import Enum, unique
@@ -22,9 +25,8 @@ _field_kernel_Gram_Kxy = 'mat_kernel_Gram_Kxy'
 
 @unique
 class ImplementedMMDSchemes(Enum):
-    GAMMA = 'Gamma Approximation'
-    PEARSON = 'Pearson Curve Moment Matching'
     SPECTRAL = 'Gram Matrix Spectrum'
+    PERMUTATION = 'Permutation'
 
 
 class MMDTestingSchemeNotImplemented(Exception):
@@ -148,7 +150,8 @@ def compute_biased_squared_mmd(data_x, data_y, kernel):
     return dic
 
 
-def perform_gram_matrix_spectrum_mmd(data_x, data_y, kernel, test_level):
+def perform_permutation_mmd(data_x, data_y, kernel, test_level,
+                            nb_permutations=5000):
 
     mmd_dic = compute_biased_squared_mmd(
         data_x=data_x,
@@ -172,18 +175,18 @@ def perform_gram_matrix_spectrum_mmd(data_x, data_y, kernel, test_level):
     mat_kernel_Gram_Kxy[0:nx, 0:nx] = mat_kernel_Gram_Kx
     mat_kernel_Gram_Kxy[nx:(2 * nx), nx:(2 * nx)] = mat_kernel_Gram_Ky
     mat_kernel_Gram_Kxy[0:nx, nx:(2 * nx)] = mat_Kxy
-    mat_kernel_Gram_Kxy[nx:(2 * nx), 0:nx] = mat_Kxy.tranpose()
+    mat_kernel_Gram_Kxy[nx:(2 * nx), 0:nx] = mat_Kxy.transpose()
 
-    mat_H = np.identity(2 * nx) - (1 / (2 * nx)) * np.ones((2 * nx, 2 * nx))
-    mat_centered_kernel_Gram_Kxy = mat_H @ mat_kernel_Gram_Kxy @ mat_H
-
-    # Simulating draws from the limit distribution
-    eigenvalues, _ = np.linalg.eig(mat_centered_kernel_Gram_Kxy)
-    eigenvalues /= (2 * nx)
-    gaussians = sqrt(2) * norm.rvs(loc=0, scale=1, size=(2 * nx))
-    draws_from_null = np.dot(
-        eigenvalues, gaussians ** 2
-    )
+    draws_from_null = np.zeros(nb_permutations)
+    permuted_matrix = copy.deepcopy(mat_kernel_Gram_Kxy)
+    for i in range(nb_permutations):
+        np.random.shuffle(permuted_matrix)
+        draws_from_null[i] = (
+            np.sum(permuted_matrix[0:nx, 0:nx]) +
+            np.sum(permuted_matrix[nx:(2 * nx), nx:(2 * nx)]) -
+            np.sum(permuted_matrix[0:nx, nx:(2 * nx)]) -
+            np.sum(permuted_matrix[nx:(2 * nx), 0:nx])
+        ) / (nx * nx)
 
     # Compute the empirical quantile
     threshold = np.quantile(draws_from_null, 1 - test_level)
@@ -197,6 +200,57 @@ def perform_gram_matrix_spectrum_mmd(data_x, data_y, kernel, test_level):
     return dic
 
 
+def perform_gram_matrix_spectrum_mmd(data_x, data_y, kernel, test_level,
+                                     nb_simulations=5000):
+
+    mmd_dic = compute_biased_squared_mmd(
+        data_x=data_x,
+        data_y=data_y,
+        kernel=kernel
+    )
+    mmd = mmd_dic['MMD']
+    mat_kernel_Gram_Kx = mmd_dic[_field_kernel_Gram_Kx]
+    mat_kernel_Gram_Ky = mmd_dic[_field_kernel_Gram_Ky]
+    mat_Kxy = mmd_dic[_field_mat_Kxy]
+
+    nx = mat_kernel_Gram_Kx.shape[0]
+    ny = mat_kernel_Gram_Ky.shape[0]
+
+    if nx != ny:
+
+        msg = 'This method is designed for equal sample sizes for X and Y.'
+        raise CaseNotHandled(msg)
+
+    mat_kernel_Gram_Kxy = np.zeros((2 * nx, 2 * nx))
+    mat_kernel_Gram_Kxy[0:nx, 0:nx] = mat_kernel_Gram_Kx
+    mat_kernel_Gram_Kxy[nx:(2 * nx), nx:(2 * nx)] = mat_kernel_Gram_Ky
+    mat_kernel_Gram_Kxy[0:nx, nx:(2 * nx)] = mat_Kxy
+    mat_kernel_Gram_Kxy[nx:(2 * nx), 0:nx] = mat_Kxy.transpose()
+
+    mat_H = np.identity(2 * nx) - (1 / (2 * nx)) * np.ones((2 * nx, 2 * nx))
+    mat_centered_kernel_Gram_Kxy = mat_H @ mat_kernel_Gram_Kxy @ mat_H
+
+    # Simulating draws from the limit distribution
+    eigenvalues = np.linalg.eigvalsh(mat_centered_kernel_Gram_Kxy)
+    eigenvalues /= (2 * nx)
+    nb_eigenvalues = eigenvalues.shape[0]
+    draws_from_null = np.zeros((nb_simulations, 1))
+    for i in range(nb_simulations):
+        gaussians = sqrt(2) * norm.rvs(loc=0, scale=1, size=nb_eigenvalues)
+        draws_from_null[i, 0] = np.dot(eigenvalues, gaussians ** 2)
+
+    # Compute the empirical quantile
+    threshold = np.quantile(draws_from_null, 1 - test_level)
+
+    dic = dict()
+    dic['Reject H0 (H0 : P_X = P_Y)'] = nx * mmd > threshold
+    dic['MMD'] = nx * mmd
+    dic['Rejection threshold'] = threshold
+    dic['Samples from simulated null'] = draws_from_null
+
+    return dic
+
+
 def perform_mmd_test(data_x, data_y, kernel, test_level, scheme):
 
     if scheme is ImplementedMMDSchemes.SPECTRAL:
@@ -210,6 +264,17 @@ def perform_mmd_test(data_x, data_y, kernel, test_level, scheme):
 
         return test_dic
 
+    # elif scheme is ImplementedMMDSchemes.PERMUTATION:
+    #
+    #     test_dic = perform_permutation_mmd(
+    #         data_x=data_x,
+    #         data_y=data_y,
+    #         kernel=kernel,
+    #         test_level=test_level
+    #     )
+    #
+    #     return test_dic
+
     else:
 
         msg = f"Scheme '{scheme}' not implemented for the MMD two-sample test."
